/*
+----------------------------------------------------------------------+
| Copyright IBM Corporation 2006.                                      |
| All Rights Reserved.                                                 |
+----------------------------------------------------------------------+
|                                                                      |
| Licensed under the Apache License, Version 2.0 (the "License"); you  |
| may not use this file except in compliance with the License. You may |
| obtain a copy of the License at                                      |
| http://www.apache.org/licenses/LICENSE-2.0                           |
|                                                                      |
| Unless required by applicable law or agreed to in writing, software  |
| distributed under the License is distributed on an "AS IS" BASIS,    |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or      |
| implied. See the License for the specific language governing         |
| permissions and limitations under the License.                       |
+----------------------------------------------------------------------+
| Authors: Dave Renshaw, Jerry Stevens                                 |
+----------------------------------------------------------------------+

   $Id: SAM_Message.c,v 1.3 2006/12/08 17:39:54 dsr Exp $

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

/* Includes for the Zend PHP API */
#include <php.h>
#include <php_ini.h>
#include <ext/standard/info.h>

/* Includes for our implementation */
#include "php_sam.h"

extern int SAM_DEBUG;

/**
  * \brief Class entry global variable for the message class.
  */
zend_class_entry * sam_ce_msg;

extern void setError(zval *objvar, int errorCode, char *errorString);

/**
  * \brief List of methods available to the PHP message object.
  *        Mapping is the PHP name of the function to the C function
  *        name. Use the the same PHP method name as the PHP class name
  *        to provide a custom constructor. Terminate this list with
  *        three NULLs.
  */
function_entry sam_message_functions[] = {
    PHP_NAMED_FE(SAMMessage, SAM_Message_ctor,    NULL)
    {NULL, NULL, NULL}
};

/***************************************************************************/
/*                                                                         */
/* Function implementations                                                */
/*                                                                         */
/***************************************************************************/

/* {{{ proto PHP_NAMED_FUNCTION(SAM_Message_ctor)
  * Main constructor function for a SAMMessage
  */
PHP_NAMED_FUNCTION(SAM_Message_ctor)
{
    char *aMsg = NULL;
    int msgLength = 0;
    zval *zParm = NULL;
    zval *zBody = NULL;

    if (SAM_DEBUG) zend_printf("<br/><b>SAM_Message->ctor()</b> enter<br/>");

    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|z", &zParm) == FAILURE) {
       setError(this_ptr, 1, "Could not parse parameters.");
       WRONG_PARAM_COUNT;
       return;
    }

    /* Store the message body property in this object's hash table */
    if (zParm != NULL) {
       MAKE_STD_ZVAL(zBody);
       ZVAL_STRINGL(zBody, Z_STRVAL_P(zParm), Z_STRLEN_P(zParm), 1);
       if (SAM_DEBUG) {
           zend_printf("<b>SAM_Message->ctor()</b> storing message body<br/>");
           switch (Z_TYPE_P(zBody)) {
              case IS_STRING:
                 if (SAM_DEBUG) zend_printf("<b>SAM_Message->ctor()</b> Message body is a string<br/>");
                 break;
              default:
                 if (SAM_DEBUG) zend_printf("<b>SAM_Message->ctor()</b> Message body is of type %d<br/>", Z_TYPE_P(zBody));
                  convert_to_string(zBody);
                 break;
           }
       }
      /* if (add_property_zval_ex(this_ptr, SAM_MESSAGE_BODY, sizeof(SAM_MESSAGE_BODY), zBody) == FAILURE) { */
         if (add_property_zval(this_ptr, SAM_MESSAGE_BODY, zBody) == FAILURE) {

           if (SAM_DEBUG) zend_printf("<b>SAM_Message->ctor()</b> storing message body FAILED!<br/>");
       }
    }

    if (SAM_DEBUG) zend_printf("<b>SAM_Message->ctor()</b> exit<br/>");
    return;
}
/* }}} */

